@@ -16,11 +16,14 @@ import serial
 from serial.tools.list_ports import comports
 from serial.tools import hexlify_codec
 
+# pylint: disable=wrong-import-order,wrong-import-position
+
 codecs.register(lambda c: hexlify_codec.getregentry() if c == 'hexlify' else None)
 
 try:
     raw_input
 except NameError:
+    # pylint: disable=redefined-builtin,invalid-name
     raw_input = input   # in python3 it's "raw"
     unichr = chr
 
@@ -36,6 +39,8 @@ def key_description(character):
 
 # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 class ConsoleBase(object):
+    """OS abstraction for console (input/output codec, no echo)"""
+
     def __init__(self):
         if sys.version_info >= (3, 0):
             self.byte_output = sys.stdout.buffer
@@ -44,20 +49,23 @@ class ConsoleBase(object):
         self.output = sys.stdout
 
     def setup(self):
-        pass
+        """Set console to read single characters, no echo"""
 
     def cleanup(self):
-        pass
+        """Restore default console settings"""
 
     def getkey(self):
+        """Read a single key from the console"""
         return None
 
-    def write_bytes(self, s):
-        self.byte_output.write(s)
+    def write_bytes(self, byte_string):
+        """Write bytes (already encoded)"""
+        self.byte_output.write(byte_string)
         self.byte_output.flush()
 
-    def write(self, s):
-        self.output.write(s)
+    def write(self, text):
+        """Write string"""
+        self.output.write(text)
         self.output.flush()
 
     #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
@@ -72,11 +80,13 @@ class ConsoleBase(object):
         self.setup()
 
 
-if os.name == 'nt':
+if os.name == 'nt':  # noqa
     import msvcrt
     import ctypes
 
     class Out(object):
+        """file-like wrapper that uses os.write"""
+
         def __init__(self, fd):
             self.fd = fd
 
@@ -145,7 +155,8 @@ elif os.name == 'posix':
             termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.old)
 
 else:
-    raise NotImplementedError("Sorry no implementation for your platform (%s) available." % sys.platform)
+    raise NotImplementedError(
+        'Sorry no implementation for your platform ({}) available.'.format(sys.platform))
 
 
 # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -190,10 +201,11 @@ class NoTerminal(Transform):
     """remove typical terminal control codes from input"""
 
     REPLACEMENT_MAP = dict((x, 0x2400 + x) for x in range(32) if unichr(x) not in '\r\n\b\t')
-    REPLACEMENT_MAP.update({
+    REPLACEMENT_MAP.update(
+        {
             0x7F: 0x2421,  # DEL
             0x9B: 0x2425,  # CSI
-            })
+        })
 
     def rx(self, text):
         return text.translate(self.REPLACEMENT_MAP)
@@ -205,11 +217,12 @@ class NoControls(NoTerminal):
     """Remove all control codes, incl. CR+LF"""
 
     REPLACEMENT_MAP = dict((x, 0x2400 + x) for x in range(32))
-    REPLACEMENT_MAP.update({
-            32: 0x2423,    # visual space
+    REPLACEMENT_MAP.update(
+        {
+            0x20: 0x2423,  # visual space
             0x7F: 0x2421,  # DEL
             0x9B: 0x2425,  # CSI
-            })
+        })
 
 
 class Printable(Transform):
@@ -217,13 +230,13 @@ class Printable(Transform):
 
     def rx(self, text):
         r = []
-        for t in text:
-            if ' ' <= t < '\x7f' or t in '\r\n\b\t':
-                r.append(t)
-            elif t < ' ':
-                r.append(unichr(0x2400 + ord(t)))
+        for c in text:
+            if ' ' <= c < '\x7f' or c in '\r\n\b\t':
+                r.append(c)
+            elif c < ' ':
+                r.append(unichr(0x2400 + ord(c)))
             else:
-                r.extend(unichr(0x2080 + ord(d) - 48) for d in '{:d}'.format(ord(t)))
+                r.extend(unichr(0x2080 + ord(d) - 48) for d in '{:d}'.format(ord(c)))
                 r.append(' ')
         return ''.join(r)
 
@@ -264,19 +277,19 @@ class DebugIO(Transform):
 # - insert newline after: a) timeout b) packet end character
 
 EOL_TRANSFORMATIONS = {
-        'crlf': CRLF,
-        'cr': CR,
-        'lf': LF,
-        }
+    'crlf': CRLF,
+    'cr': CR,
+    'lf': LF,
+}
 
 TRANSFORMATIONS = {
-        'direct': Transform,    # no transformation
-        'default': NoTerminal,
-        'nocontrol': NoControls,
-        'printable': Printable,
-        'colorize': Colorize,
-        'debug': DebugIO,
-        }
+    'direct': Transform,    # no transformation
+    'default': NoTerminal,
+    'nocontrol': NoControls,
+    'printable': Printable,
+    'colorize': Colorize,
+    'debug': DebugIO,
+}
 
 
 # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -324,6 +337,11 @@ class Miniterm(object):
         self.update_transformations()
         self.exit_character = 0x1d  # GS/CTRL+]
         self.menu_character = 0x14  # Menu: CTRL+T
+        self.alive = None
+        self._reader_alive = None
+        self.receiver_thread = None
+        self.rx_decoder = None
+        self.tx_decoder = None
 
     def _start_reader(self):
         """Start reader thread"""
@@ -339,6 +357,7 @@ class Miniterm(object):
         self.receiver_thread.join()
 
     def start(self):
+        """start worker threads"""
         self.alive = True
         self._start_reader()
         # enter console->serial loop
@@ -348,48 +367,52 @@ class Miniterm(object):
         self.console.setup()
 
     def stop(self):
+        """set flag to stop worker threads"""
         self.alive = False
 
     def join(self, transmit_only=False):
+        """wait for worker threads to terminate"""
         self.transmitter_thread.join()
         if not transmit_only:
             self.receiver_thread.join()
 
     def update_transformations(self):
-        transformations = [EOL_TRANSFORMATIONS[self.eol]] + [TRANSFORMATIONS[f] for f in self.filters]
+        """take list of transformation classes and instantiate them for rx and tx"""
+        transformations = [EOL_TRANSFORMATIONS[self.eol]] + [TRANSFORMATIONS[f]
+                                                             for f in self.filters]
         self.tx_transformations = [t() for t in transformations]
         self.rx_transformations = list(reversed(self.tx_transformations))
 
     def set_rx_encoding(self, encoding, errors='replace'):
+        """set encoding for received data"""
         self.input_encoding = encoding
         self.rx_decoder = codecs.getincrementaldecoder(encoding)(errors)
 
     def set_tx_encoding(self, encoding, errors='replace'):
+        """set encoding for transmitted data"""
         self.output_encoding = encoding
         self.tx_encoder = codecs.getincrementalencoder(encoding)(errors)
 
     def dump_port_settings(self):
+        """Write current settings to sys.stderr"""
         sys.stderr.write("\n--- Settings: {p.name}  {p.baudrate},{p.bytesize},{p.parity},{p.stopbits}\n".format(
-                p=self.serial))
+            p=self.serial))
         sys.stderr.write('--- RTS: {:8}  DTR: {:8}  BREAK: {:8}\n'.format(
-                ('active' if self.serial.rts else 'inactive'),
-                ('active' if self.serial.dtr else 'inactive'),
-                ('active' if self.serial.break_condition else 'inactive')))
+            ('active' if self.serial.rts else 'inactive'),
+            ('active' if self.serial.dtr else 'inactive'),
+            ('active' if self.serial.break_condition else 'inactive')))
         try:
             sys.stderr.write('--- CTS: {:8}  DSR: {:8}  RI: {:8}  CD: {:8}\n'.format(
-                    ('active' if self.serial.cts else 'inactive'),
-                    ('active' if self.serial.dsr else 'inactive'),
-                    ('active' if self.serial.ri else 'inactive'),
-                    ('active' if self.serial.cd else 'inactive')))
+                ('active' if self.serial.cts else 'inactive'),
+                ('active' if self.serial.dsr else 'inactive'),
+                ('active' if self.serial.ri else 'inactive'),
+                ('active' if self.serial.cd else 'inactive')))
         except serial.SerialException:
             # on RFC 2217 ports, it can happen if no modem state notification was
             # yet received. ignore this error.
             pass
         sys.stderr.write('--- software flow control: {}\n'.format('active' if self.serial.xonxoff else 'inactive'))
         sys.stderr.write('--- hardware flow control: {}\n'.format('active' if self.serial.rtscts else 'inactive'))
-        #~ sys.stderr.write('--- data escaping: %s  linefeed: %s\n' % (
-                #~ REPR_MODES[self.repr_mode],
-                #~ LF_MODES[self.convert_outgoing]))
         sys.stderr.write('--- serial input encoding: {}\n'.format(self.input_encoding))
         sys.stderr.write('--- serial output encoding: {}\n'.format(self.output_encoding))
         sys.stderr.write('--- EOL: {}\n'.format(self.eol.upper()))
@@ -495,8 +518,8 @@ class Miniterm(object):
         elif c == '\x06':                       # CTRL+F -> edit filters
             sys.stderr.write('\n--- Available Filters:\n')
             sys.stderr.write('\n'.join(
-                    '---   {:<10} = {.__doc__}'.format(k, v)
-                    for k, v in sorted(TRANSFORMATIONS.items())))
+                '---   {:<10} = {.__doc__}'.format(k, v)
+                for k, v in sorted(TRANSFORMATIONS.items())))
             sys.stderr.write('\n--- Enter new filter name(s) [{}]: '.format(' '.join(self.filters)))
             with self.console:
                 new_filters = sys.stdin.readline().lower().split()
@@ -571,7 +594,7 @@ class Miniterm(object):
                 try:
                     self.serial.baudrate = int(sys.stdin.readline().strip())
                 except ValueError as e:
-                    sys.stderr.write('--- ERROR setting baudrate: %s ---\n'.format(e))
+                    sys.stderr.write('--- ERROR setting baudrate: {} ---\n'.format(e))
                     self.serial.baudrate = backup
                 else:
                     self.dump_port_settings()
@@ -615,6 +638,7 @@ class Miniterm(object):
             sys.stderr.write('--- unknown menu character {} --\n'.format(key_description(c)))
 
     def get_help_text(self):
+        """return the help text"""
         # help text, starts with blank line!
         return """
 --- pySerial ({version}) - miniterm - help
@@ -640,149 +664,147 @@ class Miniterm(object):
 ---    b          change baud rate
 ---    x X        disable/enable software flow control
 ---    r R        disable/enable hardware flow control
-""".format(
-                version=getattr(serial, 'VERSION', 'unknown version'),
-                exit=key_description(self.exit_character),
-                menu=key_description(self.menu_character),
-                rts=key_description('\x12'),
-                dtr=key_description('\x04'),
-                brk=key_description('\x02'),
-                echo=key_description('\x05'),
-                info=key_description('\x09'),
-                upload=key_description('\x15'),
-                repr=key_description('\x01'),
-                filter=key_description('\x06'),
-                eol=key_description('\x0c'),
-                )
+""".format(version=getattr(serial, 'VERSION', 'unknown version'),
+           exit=key_description(self.exit_character),
+           menu=key_description(self.menu_character),
+           rts=key_description('\x12'),
+           dtr=key_description('\x04'),
+           brk=key_description('\x02'),
+           echo=key_description('\x05'),
+           info=key_description('\x09'),
+           upload=key_description('\x15'),
+           repr=key_description('\x01'),
+           filter=key_description('\x06'),
+           eol=key_description('\x0c'))
 
 
 # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 # default args can be used to override when calling main() from an other script
 # e.g to create a miniterm-my-device.py
 def main(default_port=None, default_baudrate=9600, default_rts=None, default_dtr=None):
+    """Command line tool, entry point"""
+
     import argparse
 
     parser = argparse.ArgumentParser(
-            description="Miniterm - A simple terminal program for the serial port.")
+        description="Miniterm - A simple terminal program for the serial port.")
 
     parser.add_argument(
-            "port",
-            nargs='?',
-            help="serial port name ('-' to show port list)",
-            default=default_port)
+        "port",
+        nargs='?',
+        help="serial port name ('-' to show port list)",
+        default=default_port)
 
     parser.add_argument(
-            "baudrate",
-            nargs='?',
-            type=int,
-            help="set baud rate, default: %(default)s",
-            default=default_baudrate)
+        "baudrate",
+        nargs='?',
+        type=int,
+        help="set baud rate, default: %(default)s",
+        default=default_baudrate)
 
     group = parser.add_argument_group("port settings")
 
     group.add_argument(
-            "--parity",
-            choices=['N', 'E', 'O', 'S', 'M'],
-            type=lambda c: c.upper(),
-            help="set parity, one of {N E O S M}, default: N",
-            default='N')
+        "--parity",
+        choices=['N', 'E', 'O', 'S', 'M'],
+        type=lambda c: c.upper(),
+        help="set parity, one of {N E O S M}, default: N",
+        default='N')
 
     group.add_argument(
-            "--rtscts",
-            action="store_true",
-            help="enable RTS/CTS flow control (default off)",
-            default=False)
+        "--rtscts",
+        action="store_true",
+        help="enable RTS/CTS flow control (default off)",
+        default=False)
 
     group.add_argument(
-            "--xonxoff",
-            action="store_true",
-            help="enable software flow control (default off)",
-            default=False)
+        "--xonxoff",
+        action="store_true",
+        help="enable software flow control (default off)",
+        default=False)
 
     group.add_argument(
-            "--rts",
-            type=int,
-            help="set initial RTS line state (possible values: 0, 1)",
-            default=default_rts)
+        "--rts",
+        type=int,
+        help="set initial RTS line state (possible values: 0, 1)",
+        default=default_rts)
 
     group.add_argument(
-            "--dtr",
-            type=int,
-            help="set initial DTR line state (possible values: 0, 1)",
-            default=default_dtr)
+        "--dtr",
+        type=int,
+        help="set initial DTR line state (possible values: 0, 1)",
+        default=default_dtr)
 
     group.add_argument(
-            "--ask",
-            action="store_true",
-            help="ask again for port when open fails",
-            default=False)
+        "--ask",
+        action="store_true",
+        help="ask again for port when open fails",
+        default=False)
 
     group = parser.add_argument_group("data handling")
 
     group.add_argument(
-            "-e", "--echo",
-            action="store_true",
-            help="enable local echo (default off)",
-            default=False)
+        "-e", "--echo",
+        action="store_true",
+        help="enable local echo (default off)",
+        default=False)
 
     group.add_argument(
-            "--encoding",
-            dest="serial_port_encoding",
-            metavar="CODEC",
-            help="set the encoding for the serial port (e.g. hexlify, Latin1, UTF-8), default: %(default)s",
-            default='UTF-8')
+        "--encoding",
+        dest="serial_port_encoding",
+        metavar="CODEC",
+        help="set the encoding for the serial port (e.g. hexlify, Latin1, UTF-8), default: %(default)s",
+        default='UTF-8')
 
     group.add_argument(
-            "-f", "--filter",
-            action="append",
-            metavar="NAME",
-            help="add text transformation",
-            default=[])
+        "-f", "--filter",
+        action="append",
+        metavar="NAME",
+        help="add text transformation",
+        default=[])
 
     group.add_argument(
-            "--eol",
-            choices=['CR', 'LF', 'CRLF'],
-            type=lambda c: c.upper(),
-            help="end of line mode",
-            default='CRLF')
+        "--eol",
+        choices=['CR', 'LF', 'CRLF'],
+        type=lambda c: c.upper(),
+        help="end of line mode",
+        default='CRLF')
 
     group.add_argument(
-            "--raw",
-            action="store_true",
-            help="Do no apply any encodings/transformations",
-            default=False)
+        "--raw",
+        action="store_true",
+        help="Do no apply any encodings/transformations",
+        default=False)
 
     group = parser.add_argument_group("hotkeys")
 
     group.add_argument(
-            "--exit-char",
-            type=int,
-            metavar='NUM',
-            help="Unicode of special character that is used to exit the application, default: %(default)s",
-            default=0x1d  # GS/CTRL+]
-            )
+        "--exit-char",
+        type=int,
+        metavar='NUM',
+        help="Unicode of special character that is used to exit the application, default: %(default)s",
+        default=0x1d)  # GS/CTRL+]
 
     group.add_argument(
-            "--menu-char",
-            type=int,
-            metavar='NUM',
-            help="Unicode code of special character that is used to control miniterm (menu), default: %(default)s",
-            default=0x14  # Menu: CTRL+T
-            )
+        "--menu-char",
+        type=int,
+        metavar='NUM',
+        help="Unicode code of special character that is used to control miniterm (menu), default: %(default)s",
+        default=0x14)  # Menu: CTRL+T
 
     group = parser.add_argument_group("diagnostics")
 
     group.add_argument(
-            "-q", "--quiet",
-            action="store_true",
-            help="suppress non-error messages",
-            default=False)
+        "-q", "--quiet",
+        action="store_true",
+        help="suppress non-error messages",
+        default=False)
 
     group.add_argument(
-            "--develop",
-            action="store_true",
-            help="show Python traceback on error",
-            default=False)
+        "--develop",
+        action="store_true",
+        help="show Python traceback on error",
+        default=False)
 
     args = parser.parse_args()
 
@@ -793,8 +815,8 @@ def main(default_port=None, default_baudrate=9600, default_rts=None, default_dtr
         if 'help' in args.filter:
             sys.stderr.write('Available filters:\n')
             sys.stderr.write('\n'.join(
-                    '{:<10} = {.__doc__}'.format(k, v)
-                    for k, v in sorted(TRANSFORMATIONS.items())))
+                '{:<10} = {.__doc__}'.format(k, v)
+                for k, v in sorted(TRANSFORMATIONS.items())))
             sys.stderr.write('\n')
             sys.exit(1)
         filters = args.filter
@@ -814,13 +836,13 @@ def main(default_port=None, default_baudrate=9600, default_rts=None, default_dtr
                     parser.error('port is not given')
         try:
             serial_instance = serial.serial_for_url(
-                    args.port,
-                    args.baudrate,
-                    parity=args.parity,
-                    rtscts=args.rtscts,
-                    xonxoff=args.xonxoff,
-                    timeout=1,
-                    do_not_open=True)
+                args.port,
+                args.baudrate,
+                parity=args.parity,
+                rtscts=args.rtscts,
+                xonxoff=args.xonxoff,
+                timeout=1,
+                do_not_open=True)
 
             if args.dtr is not None:
                 if not args.quiet:
@@ -844,10 +866,10 @@ def main(default_port=None, default_baudrate=9600, default_rts=None, default_dtr
             break
 
     miniterm = Miniterm(
-            serial_instance,
-            echo=args.echo,
-            eol=args.eol.lower(),
-            filters=filters)
+        serial_instance,
+        echo=args.echo,
+        eol=args.eol.lower(),
+        filters=filters)
     miniterm.exit_character = unichr(args.exit_char)
     miniterm.menu_character = unichr(args.menu_char)
     miniterm.raw = args.raw
@@ -856,13 +878,12 @@ def main(default_port=None, default_baudrate=9600, default_rts=None, default_dtr
 
     if not args.quiet:
         sys.stderr.write('--- Miniterm on {p.name}  {p.baudrate},{p.bytesize},{p.parity},{p.stopbits} ---\n'.format(
-                p=miniterm.serial))
+            p=miniterm.serial))
         sys.stderr.write('--- Quit: {} | Menu: {} | Help: {} followed by {} ---\n'.format(
-                key_description(miniterm.exit_character),
-                key_description(miniterm.menu_character),
-                key_description(miniterm.menu_character),
-                key_description('\x08'),
-                ))
+            key_description(miniterm.exit_character),
+            key_description(miniterm.menu_character),
+            key_description(miniterm.menu_character),
+            key_description('\x08')))
 
     miniterm.start()
     try:
